
'################################################################################################################################
' I2C Library Functions --- for initialising MPSSE and for sending/receiving
'################################################################################################################################

' NOTE:   This code is provided as an example only and is not supported or guaranteed by FTDI. The designer of the final application
' incorporating any of this sample code is responsible for ensuring it's safe and correct operation. By using this example, the designer 
' of the final application/system agrees to accept full responsibiliy for any consequences resulting from its use.
'
' Version History:
'
' Version   Date        Author      Comments
' --------- ----------- ----------- --------------------------------------------
' 1.0       Feb 2017    G Brown     Initial release


Public Module FT232H_I2C

    ' I2C Library -- Constants
    Public Const I2C_PIN_DIR As Byte = &H3
    Public Const I2C_SDAHI_SCLHI As Byte = &H7
    Public Const I2C_SDALO_SCLHI As Byte = &H5
    Public Const I2C_SDAHI_SCLLO As Byte = &H6
    Public Const I2C_SDALO_SCLLO As Byte = &H4

    Public Const MSB_FALLING_EDGE_CLOCK_BYTE_IN As Byte = &H24
    Public Const MSB_RISING_EDGE_CLOCK_BYTE_IN As Byte = &H20
    Public Const MSB_FALLING_EDGE_CLOCK_BYTE_OUT As Byte = &H11
    Public Const MSB_DOWN_EDGE_CLOCK_BIT_IN As Byte = &H26
    Public Const MSB_UP_EDGE_CLOCK_BYTE_IN As Byte = &H20
    Public Const MSB_UP_EDGE_CLOCK_BYTE_OUT As Byte = &H10
    Public Const MSB_RISING_EDGE_CLOCK_BIT_IN As Byte = &H22
    Public Const MSB_FALLING_EDGE_CLOCK_BIT_OUT As Byte = &H13

    ' I2C Library Functions -- Variables and Definitions
    Public ClockDivisor As Integer = &H31 ' dec 49, I2C rate 400KHz,     &H13 for 1MHz 

    Public GPIO_Low_Dir As Byte
    Public GPIO_Low_Dat As Byte

    ' Sending
    Public SendBufferIndex As Integer
    Public SendBuffer(0 To 255) As Byte
    Public BytesSent As Integer

    ' Receiving
    Public InputBuffer(0 To 255) As Byte
    Public BytesToRead As Byte
    Public BytesRead As Integer
    Public NumBytesInQueue As Integer
    Public ReceiveBuffer(0 To 255) As Byte
    Public ReceiveBuffer1(0 To 255) As Byte
    Public QueueTimeOut As Integer
    Public I2C_Ack As Boolean

    ' Status
    Public I2C_Status As Integer

    ' GPIO
    Public ADbusReadVal As Byte
    Public ACbusReadVal As Byte



    '#############################################################################################################################
    '#############################################################################################################################

    Public Function I2C_ConfigureMpsse() As Byte
        ' This function puts the device (which should be already open) into MPSSE mode and configures it for I2C
        ' Refer to AN_355 for further details

        '#### Reset bit mode to whatever is in the EEPROM ####
        FT_Status = FT_SetBitMode(FT_Handle, &H0, &H0)

        ' Check if the function failed...
        If FT_Status <> FT_OK Then
            Return &H1                                                                  ' Return non-zero if failed
        End If

        Sleep(10)

        '#### Set Bit mode to MPSSE Mode (2) and I/O directions to all input (0) ####
        FT_Status = FT_SetBitMode(FT_Handle, &H0, &H2)

        ' Check if the function failed...
        If FT_Status <> FT_OK Then
            Return &H1                                                                  ' Return non-zero if failed
        End If

        '#### Set Latency Timer ####

        ' Keep latency timer at 16ms, we use send immediates where we want to get reponses back quickly
        FT_Status = FT_SetLatencyTimer(FT_Handle, 16)

        ' Check if the function failed...
        If FT_Status <> FT_OK Then
            Return &H1                                                                  ' Return non-zero if failed
        End If

        '#### Set Flow Control ####

        ' RTS_CTS mode enables flow control in the driver
        FT_Status = FT_SetFlowControl(FT_Handle, FT_FLOW_RTS_CTS, 0, 0)

        ' Check if the function failed...
        If FT_Status <> FT_OK Then
            Return &H1                                                                  ' Return non-zero if failed
        End If

        '#### Set Timeouts to 5 sec ####

        ' Set timeouts to 5 sec for read and 5 sec
        FT_Status = FT_SetTimeouts(FT_Handle, 5000, 5000)

        ' Check if the function failed...
        If FT_Status <> FT_OK Then
            Return &H1                                                                  ' Return non-zero if failed
        End If

        '#### Software method of purging the buffer, by reading all data available ####
        FT_Status = FlushBuffer()

        ' Check if the function failed...
        If FT_Status <> FT_OK Then
            Return &H1                                                                  ' Return non-zero if failed
        End If

        '#### Sync 1 - Send an invalid command 0xAA check that MPSSE replies with 'Invalid Command' ####

        SendBuffer(0) = &HAA                                                            ' 0xAA is an invalid command for the MPSSE

        I2C_Status = Send_Data(1)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        End If

        ' Now wait for a reply- MPSSE will return 0xFA followed by the bad command 0xAA 

        I2C_Status = Receive_Data(2)

        If (I2C_Status <> 0) Then
            Return &H1
        End If

        ' Check that we got the correct data back otherwise cancel the initialisation
        If ((ReceiveBuffer(0) <> &HFA) Or (ReceiveBuffer(1) <> &HAA)) Then
            Return &H1                                                              ' Return non-zero if failed
        End If


        '######## Sync 2 - Send an invalid command 0xAB check that MPSSE replies with 'Invalid Command' ########

        ' 0xAB is an invalid command for the MPSSE
        SendBuffer(0) = &HAB

        I2C_Status = Send_Data(1)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        End If

        ' Now wait for a reply - MPSSE will return 0xFA followed by the bad command 0xAB 

        I2C_Status = Receive_Data(2)

        If (I2C_Status <> 0) Then
            Return &H1
        End If

        ' Check that we got the correct data back otherwise cancel the initialisation
        If ((ReceiveBuffer(0) <> &HFA) Or (ReceiveBuffer(1) <> &HAB)) Then
            Return &H1                                                              ' Return non-zero if failed
        End If

        '######## Set GPIO to the required line states for idle I2C ########

        GPIO_Low_Dat = 0
        GPIO_Low_Dir = 0
        I2C_SetLineStatesIdle()
        ' Note:  Could also set high byte GPIO with a call here to I2C_SetGPIOValuesHigh(0,0)

        '************************************************************************************************************
        ' MPSSE Configuration
        '************************************************************************************************************

        ' SCL clock frequency can be worked out by below algorithm with divide by 5 set as off
        ' SCL frequency  = 60MHz /((1 +  [(1 +0xValueH*256) OR 0xValueL])*2)

        SendBufferIndex = 0
        SendBuffer(SendBufferIndex) = &H85                      ' MPSSE command to disable loopback
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H8A                      ' Disable clock divide by 5 for 60Mhz master clock
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H97                      ' Adaptive clocking off
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H8C                      ' Enable 3 phase data clock
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H9E                      ' Enable the FT232H's drive-zero mode with the following enable mask...
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H7                       ' ... Low byte (ADx) enables - bits 0, 1 and 2... and ...
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                       ' ...High byte (ACx) enables - all off
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H86                      ' Command to set clock divisor with values of ...
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ClockDivisor And &HFF     ' ... 0xValueL of clock divisor... and ...
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = (ClockDivisor >> 8) And &HFF      ' ... 0xValueH of clock divisor
        SendBufferIndex += 1

        I2C_Status = Send_Data(SendBufferIndex)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        End If

        Return &H0                                            ' Otherwise, return 'success' code 0

    End Function

    Public Function I2C_ReadByte(ByRef ACK As Boolean) As Byte

        Dim ADbusVal As Byte
        Dim ADbusDir As Byte
        Dim SendBufferIndex As Byte

        SendBufferIndex = 0                                                     ' Clear output buffer

        '######## Build up the buffer of commands ########

        ' Clock one byte of data in
        SendBuffer(SendBufferIndex) = MSB_RISING_EDGE_CLOCK_BYTE_IN             ' clock data byte in
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                                       '
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                                       ' Data length of 0x0000 means 1 byte data to clock in
        SendBufferIndex += 1
        ' Now send the ACK/NAK by clocking out one bit...send 0 for ACK and 1 for NAK
        SendBuffer(SendBufferIndex) = MSB_FALLING_EDGE_CLOCK_BIT_OUT            ' clock data bit out
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                                       ' Length of 0 means 1 bit
        SendBufferIndex += 1
        If (ACK = True) Then
            SendBuffer(SendBufferIndex) = &H0                                   ' Data bit to send is a 0
        Else
            SendBuffer(SendBufferIndex) = &HFF                                  ' Data bit to send is a 1
        End If
        SendBufferIndex += 1

        ADbusVal = I2C_SDAHI_SCLLO Or (GPIO_Low_Dat And &HF8)                   ' Don't change the lower 3 bits (SCK and SDA)
        ADbusDir = I2C_PIN_DIR Or (GPIO_Low_Dir And &HF8)

        ' Put line back to idle (data released, clock pulled low)
        SendBuffer(SendBufferIndex) = &H80                                      ' Command - set low byte
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusVal                                  ' Set the values
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusDir                                  ' Set the directions
        SendBufferIndex += 1
        ' This command then tells the MPSSE to send any results gathered back immediately
        SendBuffer(SendBufferIndex) = &H87                                      ' Send answer back immediate command
        SendBufferIndex += 1

        '######## Send the commands to the chip ########
        I2C_Status = Send_Data(SendBufferIndex)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        End If

        '######## Read the reply and check that the correct number of bytes came back ########

        I2C_Status = Receive_Data(1)                                                         ' Call the receive function

        If (I2C_Status <> 0) Then
            Return &H1
        End If

        InputBuffer(0) = ReceiveBuffer(0)                                   ' followed by the data itself
        Return &H0

    End Function

    Public Function I2C_Read2BytesWithAddr(ByRef Address As Byte) As Byte

        Dim ADbusVal As Byte
        Dim ADbusDir As Byte
        Dim SendBufferIndex As Byte


        Address = (Address << 1)                                                ' change 7-bit address to 8-bits with R/W bit 0 for reading
        Address = Address Or &H1

        SendBufferIndex = 0                                                     ' Clear output buffer

        '######## Send the I2C Address with read bit ########

        ' Clock one byte of data out
        SendBuffer(SendBufferIndex) = MSB_FALLING_EDGE_CLOCK_BYTE_OUT           ' clock data byte out
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                                       '
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                                       ' Data length of 0x0000 means 1 byte data to clock in
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = Address                                  ' Byte to send
        SendBufferIndex += 1

        ADbusVal = I2C_SDAHI_SCLLO Or (GPIO_Low_Dat And &HF8)                   ' Don't change the lower 3 bits (SCK and SDA)
        ADbusDir = I2C_PIN_DIR Or (GPIO_Low_Dir And &HF8)

        ' Put line back to idle (data released, clock pulled low)
        SendBuffer(SendBufferIndex) = &H80                                      ' Command - set low byte
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusVal                                  ' Set the values
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusDir                                  ' Set the directions
        SendBufferIndex += 1

        ' CLOCK IN ACK
        SendBuffer(SendBufferIndex) = MSB_RISING_EDGE_CLOCK_BIT_IN              ' clock data byte in
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                                       ' Length of 0 means 1 bit
        SendBufferIndex += 1

        '######## Read first data byte ########

        ' Clock one byte of data in
        SendBuffer(SendBufferIndex) = MSB_RISING_EDGE_CLOCK_BYTE_IN             ' clock data byte in
        SendBufferIndex += 1

        SendBuffer(SendBufferIndex) = &H0                                       '
        SendBufferIndex += 1

        SendBuffer(SendBufferIndex) = &H0                                       ' Data length of 0x0000 means 1 byte data to clock in
        SendBufferIndex += 1

        ' Now send the ACK by clocking out one bit...send ACK bit as '0'
        SendBuffer(SendBufferIndex) = MSB_FALLING_EDGE_CLOCK_BIT_OUT            ' clock data bit out
        SendBufferIndex += 1

        SendBuffer(SendBufferIndex) = &H0                                       ' Length of 0 means 1 bit
        SendBufferIndex += 1

        SendBuffer(SendBufferIndex) = &H0                                       ' Data bit to send is a '0'
        SendBufferIndex += 1


        ADbusVal = I2C_SDAHI_SCLLO Or (GPIO_Low_Dat And &HF8)                   ' Don't change the lower 3 bits (SCK and SDA)
        ADbusDir = I2C_PIN_DIR Or (GPIO_Low_Dir And &HF8)

        SendBuffer(SendBufferIndex) = &H80                                      ' Command - set low byte
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusVal                                  ' Set the values
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusDir                                  ' Set the directions
        SendBufferIndex += 1

        '######## Read second data byte ########

        ' Clock one byte of data in
        SendBuffer(SendBufferIndex) = MSB_RISING_EDGE_CLOCK_BYTE_IN             ' clock data byte in
        SendBufferIndex += 1

        SendBuffer(SendBufferIndex) = &H0                                       '
        SendBufferIndex += 1

        SendBuffer(SendBufferIndex) = &H0                                       ' Data length of 0x0000 means 1 byte data to clock in
        SendBufferIndex += 1

        ' Now send the ACK by clocking out one bit...send ACK bit as '1'
        SendBuffer(SendBufferIndex) = MSB_FALLING_EDGE_CLOCK_BIT_OUT            ' clock data bit out
        SendBufferIndex += 1

        SendBuffer(SendBufferIndex) = &H0                                       ' Length of 0 means 1 bit
        SendBufferIndex += 1

        SendBuffer(SendBufferIndex) = &HFF                                      ' Data bit to send is a '1'
        SendBufferIndex += 1


        ADbusVal = I2C_SDAHI_SCLLO Or (GPIO_Low_Dat And &HF8)                   ' Don't change the lower 3 bits (SCK and SDA)
        ADbusDir = I2C_PIN_DIR Or (GPIO_Low_Dir And &HF8)

        ' Put line back to idle (data released, clock pulled low)
        SendBuffer(SendBufferIndex) = &H80                                      ' Command - set low byte
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusVal                                  ' Set the values
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusDir                                  ' Set the directions
        SendBufferIndex += 1

        ' This command then tells the MPSSE to send any results gathered back immediately
        SendBuffer(SendBufferIndex) = &H87                                      ' Send answer back immediate command
        SendBufferIndex += 1

        '######## Send the commands to the chip ########
        I2C_Status = Send_Data(SendBufferIndex)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        End If

        '######## Read the reply and check that the correct number of bytes came back ########

        I2C_Status = Receive_Data(3)                                                         ' Call the receive function

        If (I2C_Status <> 0) Then
            Return &H1
        End If

        If ((ReceiveBuffer(0) And &H1) = &H0) Then
            I2C_Ack = True
        Else
            I2C_Ack = False
        End If
        InputBuffer(0) = ReceiveBuffer(1)                                   ' followed by the data itself
        InputBuffer(1) = ReceiveBuffer(2)                                   ' followed by the data itself
        Return &H0

    End Function

    Public Function I2C_SendDeviceAddrAndCheckACK(ByRef Address As Byte, ByRef Read As Boolean) As Byte

        Dim ADbusVal As Byte
        Dim ADbusDir As Byte
        Dim SendBufferIndex As Byte

        Address = (Address << 1)
        If (Read = True) Then
            Address = Address Or &H1
        End If

        SendBufferIndex = 0                                                     ' Clear output buffer

        ' Clock one byte of data out
        SendBuffer(SendBufferIndex) = MSB_FALLING_EDGE_CLOCK_BYTE_OUT           ' clock data byte out
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                                       '
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                                       ' Data length of 0x0000 means 1 byte data to clock in
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = Address                                  ' Byte to send
        SendBufferIndex += 1

        ADbusVal = I2C_SDAHI_SCLLO Or (GPIO_Low_Dat And &HF8)                   ' Don't change the lower 3 bits (SCK and SDA)
        ADbusDir = I2C_PIN_DIR Or (GPIO_Low_Dir And &HF8)

        ' Put line back to idle (data released, clock pulled low)
        SendBuffer(SendBufferIndex) = &H80                                      ' Command - set low byte
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusVal                                  ' Set the values
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusDir                                  ' Set the directions
        SendBufferIndex += 1

        ' CLOCK IN ACK
        SendBuffer(SendBufferIndex) = MSB_RISING_EDGE_CLOCK_BIT_IN              ' clock data byte in
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                                       ' Length of 0 means 1 bit
        SendBufferIndex += 1

        ' This command then tells the MPSSE to send any results gathered back immediately
        SendBuffer(SendBufferIndex) = &H87                                  ' Send answer back immediate command
        SendBufferIndex += 1

        '######## Send the commands to the chip ########
        I2C_Status = Send_Data(SendBufferIndex)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        End If

        '######## Read the reply and check the ACK bit ########

        I2C_Status = Receive_Data(1)                                                         ' Call the receive function

        If (I2C_Status <> 0) Then
            Return &H1
        End If

        If ((ReceiveBuffer(0) And &H1) = &H0) Then
            I2C_Ack = True
        Else
            I2C_Ack = False
        End If

        Return &H0

    End Function





    Public Function I2C_SendByteAndCheckACK(ByRef DataSend As Byte) As Byte

        Dim ADbusVal As Byte
        Dim ADbusDir As Byte
        Dim SendBufferIndex As Byte

        SendBufferIndex = 0                                                     ' Clear output buffer

        ' Clock one byte of data out
        SendBuffer(SendBufferIndex) = MSB_FALLING_EDGE_CLOCK_BYTE_OUT           ' clock data byte out
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                                       '
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                                       ' Data length of 0x0000 means 1 byte data to clock in
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = DataSend                                  ' Byte to send
        SendBufferIndex += 1

        ADbusVal = I2C_SDAHI_SCLLO Or (GPIO_Low_Dat And &HF8)                   ' Don't change the lower 3 bits (SCK and SDA)
        ADbusDir = I2C_PIN_DIR Or (GPIO_Low_Dir And &HF8)

        ' Put line back to idle (data released, clock pulled low)
        SendBuffer(SendBufferIndex) = &H80                                      ' Command - set low byte
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusVal                                  ' Set the values
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusDir                                  ' Set the directions
        SendBufferIndex += 1

        ' CLOCK IN ACK
        SendBuffer(SendBufferIndex) = MSB_RISING_EDGE_CLOCK_BIT_IN              ' clock data byte in
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H0                                       ' Length of 0 means 1 bit
        SendBufferIndex += 1

        ' This command then tells the MPSSE to send any results gathered back immediately
        SendBuffer(SendBufferIndex) = &H87                                  ' Send answer back immediate command
        SendBufferIndex += 1

        '######## Send the commands to the chip ########
        I2C_Status = Send_Data(SendBufferIndex)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        End If

        '######## Read the reply and check the ACK bit ########

        I2C_Status = Receive_Data(1)                                                         ' Call the receive function

        If (I2C_Status <> 0) Then
            Return &H1
        End If

        If ((ReceiveBuffer(0) And &H1) = &H0) Then
            I2C_Ack = True
        Else
            I2C_Ack = False
        End If

        Return &H0

    End Function


    Public Function I2C_SetStart() As Byte

        Dim ADbusVal As Byte
        Dim ADbusDir As Byte
        Dim SendBufferIndex As Byte

        SendBufferIndex = 0

        ' Idle state first
        ADbusDir = I2C_PIN_DIR Or (GPIO_Low_Dir And &HF8)
        ADbusVal = I2C_SDAHI_SCLHI Or (GPIO_Low_Dat And &HF8)
        For Counter As Byte = 0 To 6
            SendBuffer(SendBufferIndex) = &H80                      ' Command - set low byte
            SendBufferIndex += 1
            SendBuffer(SendBufferIndex) = ADbusVal                  ' Set the values 
            SendBufferIndex += 1
            SendBuffer(SendBufferIndex) = ADbusDir                  ' Set the directions
            SendBufferIndex += 1
        Next

        ' Bring data line low
        ADbusVal = I2C_SDALO_SCLHI Or (GPIO_Low_Dat And &HF8)           ' 

        For Counter As Byte = 0 To 6
            SendBuffer(SendBufferIndex) = &H80                      ' Command - set low byte
            SendBufferIndex += 1
            SendBuffer(SendBufferIndex) = ADbusVal                  ' Set the values  
            SendBufferIndex += 1
            SendBuffer(SendBufferIndex) = ADbusDir                  ' Set the directions
            SendBufferIndex += 1
        Next

        ' Bring clock line low too
        ADbusVal = I2C_SDALO_SCLLO Or (GPIO_Low_Dat And &HF8)

        For Counter As Byte = 0 To 6
            SendBuffer(SendBufferIndex) = &H80                      ' Command - set low byte
            SendBufferIndex += 1
            SendBuffer(SendBufferIndex) = ADbusVal                  ' Set the values 
            SendBufferIndex += 1
            SendBuffer(SendBufferIndex) = ADbusDir                  ' Set the directions
            SendBufferIndex += 1
        Next

        '######## Send the commands to the chip ########
        I2C_Status = Send_Data(SendBufferIndex)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        Else
            Return &H0
        End If

    End Function

    Public Function I2C_SetStop() As Byte

        Dim ADbusVal As Byte
        Dim ADbusDir As Byte
        Dim SendBufferIndex As Byte

        '######## Combine the I2C line state with the GPIO ########
        ADbusDir = I2C_PIN_DIR Or (GPIO_Low_Dir And &HF8)

        SendBufferIndex = 0

        ' Bring clock high
        ADbusVal = I2C_SDALO_SCLHI Or (GPIO_Low_Dat And &HF8)

        For Counter As Byte = 0 To 6
            SendBuffer(SendBufferIndex) = &H80                       ' Command - set low byte
            SendBufferIndex += 1
            SendBuffer(SendBufferIndex) = ADbusVal                   ' Set the values 
            SendBufferIndex += 1
            SendBuffer(SendBufferIndex) = ADbusDir                   ' Set the directions
            SendBufferIndex += 1
        Next

        ' Bring data high too
        ADbusVal = I2C_SDAHI_SCLHI Or (GPIO_Low_Dat And &HF8)

        For Counter As Byte = 0 To 6
            SendBuffer(SendBufferIndex) = &H80                       ' Command - set low byte
            SendBufferIndex += 1
            SendBuffer(SendBufferIndex) = ADbusVal                   ' Set the values
            SendBufferIndex += 1
            SendBuffer(SendBufferIndex) = ADbusDir                   ' Set the directions
            SendBufferIndex += 1
        Next

        '######## Send the commands to the chip ########
        I2C_Status = Send_Data(SendBufferIndex)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        Else
            Return &H0
        End If

    End Function

    Public Function I2C_SetLineStatesIdle() As Byte
        ' This function is used to write the I2C lines to their default idle states (i.e. when NOT within a message transaction)
        ' It is used as a GPIO write function.

        Dim SendBufferIndex As Byte
        Dim ADbusVal As Byte
        Dim ADbusDir As Byte

        '######## Combine the I2C line state with the GPIO ########
        ADbusVal = I2C_SDAHI_SCLHI Or (GPIO_Low_Dat And &HF8)       ' Don't change the lower 3 bits (SCK and SDA)
        ADbusDir = I2C_PIN_DIR Or (GPIO_Low_Dir And &HF8)

        '######## Send the command to set GPIOs ########
        SendBufferIndex = 0
        SendBuffer(SendBufferIndex) = &H80                          ' Command - set low byte
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusVal                      ' Set the values
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ADbusDir                      ' Set the directions
        SendBufferIndex += 1

        '######## Send the commands to the chip ########
        I2C_Status = Send_Data(SendBufferIndex)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        Else
            Return &H0
        End If

    End Function

    Public Function I2C_GetGPIOValuesLow() As Byte
        ' Gets the value of the bits 0-7 of the lower byte of the port (ADbus)

        Dim SendBufferIndex As Byte

        '######## Send the command to read GPIOs ########
        SendBufferIndex = 0
        SendBuffer(SendBufferIndex) = &H81                          ' Command - read low byte (== ADbus)
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H87                          ' Send result back immediately
        SendBufferIndex += 1

        '######## Send the commands to the chip ########
        I2C_Status = Send_Data(SendBufferIndex)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        End If

        '######## Read the reply and check that the correct number of bytes came back ########

        I2C_Status = Receive_Data(1)                                                         ' Call the receive function

        If (I2C_Status <> 0) Then
            Return &H1
        End If

        ADbusReadVal = ReceiveBuffer(0) And &HF8

        Return 0

    End Function


    Public Function I2C_SetGPIOValuesHigh(ByRef ACbusDir As Byte, ByRef ACbusVal As Byte) As Byte
        ' This function is used to write the ACbus lines GPIO state

        Dim SendBufferIndex As Byte

        '######## Send the command to set GPIOs ########
        SendBufferIndex = 0
        SendBuffer(SendBufferIndex) = &H82                          ' Command - set high byte
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ACbusVal                      ' Set the values
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = ACbusDir                      ' Set the directions
        SendBufferIndex += 1

        '######## Send the commands to the chip ########
        I2C_Status = Send_Data(SendBufferIndex)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        Else
            Return &H0
        End If

    End Function


    Public Function I2C_GetGPIOValuesHigh() As Byte
        ' Gets the value of the bits 0-7 of the upper byte of the port (ACbus)

        Dim SendBufferIndex As Byte

        '######## Send the command to read GPIOs ########
        SendBufferIndex = 0
        SendBuffer(SendBufferIndex) = &H83                          ' Command - read high byte (== ACbus)
        SendBufferIndex += 1
        SendBuffer(SendBufferIndex) = &H87                          ' Send result back immediately
        SendBufferIndex += 1

        '######## Send the commands to the chip ########
        I2C_Status = Send_Data(SendBufferIndex)

        ' If the write failed, return an error code 
        If (I2C_Status <> 0) Then
            Return &H1
        End If

        '######## Read the reply and check that the correct number of bytes came back ########

        I2C_Status = Receive_Data(1)                                                         ' Call the receive function

        If (I2C_Status <> 0) Then
            Return &H1
        End If

        ACbusReadVal = ReceiveBuffer(0)

        Return 0

    End Function


    '################################################################################################################################
    ' D2xx Functions
    '################################################################################################################################


    Private Function Receive_Data(ByVal BytesToRead As Integer) As Byte
        ' This function reads the specified number of bytes via FT_Read and checks for successful completion
        ' In most cases, the requested bytes will all be there almost immediately but this function has additional
        ' coding to allow for cases where the bytes are delayed by a heavily loaded USB link etc.

        Dim NumBytesToRead As Integer
        Dim NumBytesRead As Integer
        Dim BufferIndex As Byte
        Dim QueueTimeoutFlag As Boolean

        NumBytesInQueue = 0
        QueueTimeOut = 0
        BufferIndex = 0
        QueueTimeOut = 0
        QueueTimeoutFlag = False

        BufferIndex = 0

        NumBytesToRead = BytesToRead

        ' Keep looping until requested number of bytes are received or we've tried 50000 times
        ' (the number of re-tries can be fine tuned if needed)
        While ((BufferIndex < NumBytesToRead) And (QueueTimeoutFlag = False))

            ' Check how many bytes are in the driver buffer
            FT_Status = FT_GetQueueStatus(FT_Handle, NumBytesInQueue)

            If FT_Status <> 0 Then
                Return &H1
            End If

            ' If the GetQueueStatus shows any bytes available, go on to read them
            If (NumBytesInQueue > 0) Then

                FT_Status = FT_Read_Bytes(FT_Handle, ReceiveBuffer(BufferIndex), NumBytesInQueue, NumBytesRead)

                If FT_Status <> 0 Then
                    Return &H1
                End If

                If (NumBytesRead = NumBytesInQueue) Then
                    BufferIndex = BufferIndex + NumBytesRead
                Else
                    Return &H1
                End If

            End If

            ' Increment the software timeout counter. Check if the count has reached the pre-set value.
            ' The value of 50000 can be adjusted if required depending on the application.
            ' Normally, data will come back almost immediately and well within this time
            QueueTimeOut = QueueTimeOut + 1
            If (QueueTimeOut = 500000) Then
                QueueTimeoutFlag = True
            Else
                Sleep(0)
                Application.DoEvents()
            End If

        End While

        ' Set the global variable to tell the calling function how many bytes were actually read
        ' Calling function can read the bytes from ReceiveBuffer()
        BytesRead = BufferIndex

        ' Check that we got required number of bytes
        If BytesRead <> BytesToRead Then
            Return &H1      ' Error code - calling application should check BytesRead to see how many were actually received
            ' and take appropriate action
        Else
            Return &H0      ' Indicate success
        End If

    End Function

    Private Function Send_Data(ByVal BytesToSend As Integer) As Byte
        ' This function writes the specified number of bytes via FT_Write and checks for successful completion

        Dim numBytesToSend As Integer
        Dim numBytesSent As Integer

        numBytesToSend = BytesToSend

        FT_Status = FT_Write_Bytes(FT_Handle, SendBuffer(0), numBytesToSend, numBytesSent)     ' Write buffer to the device

        BytesSent = numBytesSent        ' Global variable will indicate number of bytes actually sent

        ' Check that the status was OK and that all bytes got sent.
        '       We return code 0 to the calling function if the bytes all got sent and no errors.
        '       We return nonZero value if an error occurred or not all bytes sent (timed out etc.). In this case, the calling
        '           function can check global variable BytesSent to see how many were actually written.
        If ((FT_Status <> 0) Or (numBytesToSend <> numBytesSent)) Then
            Return &H1                  ' Could extend the set of return values to indicate why it failed etc. 
        Else
            Return &H0
        End If

    End Function

    Private Function FlushBuffer() As Byte

        Dim PurgeBytes As Integer
        Dim PurgeBuffer(10000) As Byte
        Dim PurgeBytesRead As Integer

        FT_Status = FT_GetQueueStatus(FT_Handle, PurgeBytes)
        If (FT_Status <> 0) Then
            Return &H1
        End If

        If (PurgeBytes > 0) Then
            FT_Status = FT_Read_Bytes(FT_Handle, PurgeBuffer(0), PurgeBytes, PurgeBytesRead)
            If ((FT_Status = 0) And (PurgeBytes = PurgeBytesRead)) Then
                Return 0
            Else
                Return &H1
            End If
        End If

        Return 0


    End Function


    '################################################################################################################################
    ' End of    I2C Library Functions --- for initialising MPSSE and for sending/receiving
    '################################################################################################################################


End Module






















